/*******************************************************************************
 mod_unified_s3_auth.cpp

 mod_unified_s3_auth - An Apache module for S3 Authentication

 Copyright (C) 2020-2024 CodeShop B.V.

 For licensing see the LICENSE file
******************************************************************************/

#include <httpd.h>
#include <http_core.h>
#include <http_config.h>
#include <http_protocol.h>
#include <http_log.h>
#include <http_request.h>
#include <apr_version.h>
#include <apr_strings.h>
#include <apr_base64.h>
#include <apr_md5.h>

#include <mod_streaming_export.h>
#include <s3_util.h>

#include <ap_dump.h>
#include <ap_log.h>

typedef struct
{
  char* s3_secret_key;
  char* s3_access_key;
  char* s3_region;
  char* s3_security_token;
  int s3_use_headers;
} dir_config_t;

static void* create_dir_config(apr_pool_t* pool, char* dummy);
static void* merge_dir_config(apr_pool_t* pool, void* basev, void* addv);
static void register_hooks(apr_pool_t* p);

static command_rec const s3_auth_cmds[] =
{
  AP_INIT_TAKE1("S3SecretKey",
    reinterpret_cast<cmd_func>(ap_set_string_slot),
    reinterpret_cast<void*>(APR_OFFSETOF(dir_config_t, s3_secret_key)),
    ACCESS_CONF,
    "Pass the Amazon S3 Secret Key"),

  AP_INIT_TAKE1("S3AccessKey",
    reinterpret_cast<cmd_func>(ap_set_string_slot),
    reinterpret_cast<void*>(APR_OFFSETOF(dir_config_t, s3_access_key)),
    ACCESS_CONF,
    "Pass the Amazon S3 Access Key)"),

  AP_INIT_TAKE1("S3Region",
    reinterpret_cast<cmd_func>(ap_set_string_slot),
    reinterpret_cast<void*>(APR_OFFSETOF(dir_config_t, s3_region)),
    ACCESS_CONF,
    "Pass the Amazon S3 Region"),

  AP_INIT_TAKE1("S3SecurityToken",
    reinterpret_cast<cmd_func>(ap_set_string_slot),
    reinterpret_cast<void*>(APR_OFFSETOF(dir_config_t, s3_security_token)),
    ACCESS_CONF,
    "Pass the Amazon S3 Security Token"),

  AP_INIT_FLAG("S3UseHeaders",
    reinterpret_cast<cmd_func>(ap_set_flag_slot),
    reinterpret_cast<void*>(APR_OFFSETOF(dir_config_t, s3_use_headers)),
    ACCESS_CONF,
    "Set to 'on' to use headers for Amazon S3 authentication"),

  { NULL }
};

extern "C" {

module AP_MODULE_DECLARE_DATA unified_s3_auth_module =
{
  STANDARD20_MODULE_STUFF,
  create_dir_config,
  merge_dir_config,
  NULL,
  NULL,
  s3_auth_cmds,
  register_hooks
};

} // extern C definitions

static void* create_dir_config(apr_pool_t* pool, char* dummy)
{
  TRACE3_P(pool, "create_dir_config: enter, dummy=%s", dummy);
  auto* config = static_cast<dir_config_t*>(
    apr_pcalloc(pool, sizeof(dir_config_t)));
  config->s3_secret_key = NULL;
  config->s3_access_key = NULL;
  config->s3_region = NULL;
  config->s3_security_token = NULL;
  config->s3_use_headers = -1;

  TRACE3_P(pool, "create_dir_config: leave, config=%p", config);
  return config;
}

static void* merge_dir_config(apr_pool_t* pool, void* basev, void* addv)
{
  TRACE3_P(pool, "merge_dir_config: enter, basev=%p, addv=%p", basev, addv);
  auto* base = static_cast<dir_config_t*>(basev);
  auto* add = static_cast<dir_config_t*>(addv);
  auto* res = static_cast<dir_config_t*>(
    apr_pcalloc(pool, sizeof(dir_config_t)));

  res->s3_secret_key = add->s3_secret_key == NULL ?
                       base->s3_secret_key : add->s3_secret_key;
  res->s3_access_key = add->s3_access_key == NULL ?
                       base->s3_access_key : add->s3_access_key;
  res->s3_region = add->s3_region == NULL ?
                   base->s3_region : add->s3_region;
  res->s3_security_token = add->s3_security_token == NULL ?
                           base->s3_security_token : add->s3_security_token;
  res->s3_use_headers = add->s3_use_headers == -1 ?
                        base->s3_use_headers : add->s3_use_headers;
  TRACE1_P(pool, "merge_dir_config: s3_secret_key=%s, s3_access_key=%s, "
                 "s3_region=%s, s3_security_token=%s, s3_use_headers=%d",
                 res->s3_secret_key, res->s3_access_key, res->s3_region,
                 res->s3_security_token, res->s3_use_headers);

  TRACE3_P(pool, "merge_dir_config: leave, res=%p", res);
  return res;
}

static int s3_auth_translate(request_rec* r)
{
  DEBUG_R(r, "s3_auth_translate: enter, subreq=%s hostname=%s uri=%s filename=%s"
    " r=%s", (r->main == NULL ? "no" : "yes"), r->hostname, r->uri, r->filename,
    dump_request_rec(r->pool, r));

  if(r->filename == NULL)
  {
    DEBUG_R(r, "Skipping S3 auth, filename is NULL");
    return DECLINED;
  }

  auto* dir_conf = static_cast<dir_config_t*>(
    ap_get_module_config(r->per_dir_config, &unified_s3_auth_module));
  if(dir_conf->s3_secret_key == NULL || dir_conf->s3_access_key == NULL)
  {
    DEBUG_R(r, "Skipping S3 auth, no secret or access keys");
    return DECLINED;
  }

  if(dir_conf->s3_use_headers > 0)
  {
    DEBUG_R(r, "Skipping S3 auth, S3UseHeaders is on");
    return DECLINED;
  }

  DEBUG_R(r, "Adding S3 arguments, url=%s, region=%s",
    r->filename, dir_conf->s3_region);

  TRACE3_R(r, "s3_auth_translate: leave, DECLINED");
  return DECLINED;
}

static void set_header_in(void* p, char const* key, char const* value)
{
  request_rec* r = (request_rec*)p;
  DEBUG_R(r, "set header=%s, value=%s", key, value);
  apr_table_set(r->headers_in, key, value);
}

static void set_url(void *p, char const* url)
{
  request_rec* r = (request_rec*)p;
  DEBUG_R(r, "set url=proxy:%s", url);
  r->filename = apr_pstrcat(r->pool, "proxy:", url, NULL);
}

static int s3_auth_fixups(request_rec* r)
{
  DEBUG_R(r, "s3_auth_fixups: enter, subreq=%s hostname=%s uri=%s filename=%s args=%s",
    (r->main == NULL ? "no" : "yes"), r->hostname, r->uri, r->filename, r->args);

  if(r->filename == NULL)
  {
    DEBUG_R(r, "Skipping S3 auth, filename is NULL");
    return DECLINED;
  }

  static char const ismproxy[] = "ismproxy:";
  if(strncmp(r->filename, ismproxy, sizeof ismproxy - 1) == 0)
  {
    DEBUG_R(r, "Skipping S3 auth for ismproxy URL %s", r->filename);
    return DECLINED;
  }

  static char const prefix_proxy[] = "proxy:";
  if(strncmp(r->filename, prefix_proxy, sizeof prefix_proxy - 1) != 0)
  {
    DEBUG_R(r, "Skipping S3 auth for non-proxy URL %s", r->filename);
    return DECLINED;
  }
  char const* url = r->filename + sizeof prefix_proxy - 1;
  DEBUG_R(r, "proxy url=%s", url);

  static char const prefix_http[] = "http://";
  static char const prefix_https[] = "https://";
  if(strncmp(url, prefix_http, sizeof prefix_http - 1) != 0 &&
     strncmp(url, prefix_https, sizeof prefix_https - 1) != 0)
  {
    DEBUG_R(r, "Skipping S3 auth for non-http/https proxy URL %s", r->filename);
    return DECLINED;
  }

  auto* dir_conf = static_cast<dir_config_t*>(
    ap_get_module_config(r->per_dir_config, &unified_s3_auth_module));
  if(dir_conf->s3_secret_key == NULL || dir_conf->s3_access_key == NULL)
  {
    // If the request itself does not have any S3 parameters, try again with the
    // configuration of the parent request, if there is one.
    if(r->main != NULL)
    {
      dir_conf = static_cast<dir_config_t*>(
        ap_get_module_config(r->main->per_dir_config, &unified_s3_auth_module));
      if(dir_conf->s3_secret_key == NULL || dir_conf->s3_access_key == NULL)
      {
        DEBUG_R(r, "Skipping S3 auth, no secret or access keys (sub & main)");
        return DECLINED;
      }
      DEBUG_R(r, "Using S3 auth config from parent request");
    }
    else
    {
      DEBUG_R(r, "Skipping S3 auth, no secret or access keys");
      return DECLINED;
    }
  }

  if(dir_conf->s3_use_headers <= 0)
  {
    // If we need to add arguments, check first whether there is already an
    // Authorization header, and bail out in that case. AWS does not accept
    // multiple authentication mechanisms simultaneously.
    char const* auth_header = apr_table_get(r->headers_in, "Authorization");
    if(auth_header != NULL)
    {
      WARNING_R(r, "Not adding S3 authentication query arguments, because an "
        "Authorization header is already set in the request: %s", auth_header);
      return DECLINED;
    }

    DEBUG_R(r, "Adding S3 arguments, url=%s, args=%s, region=%s",
      url, r->args, dir_conf->s3_region);

    char result_text[256] = { 0 };
    int status = mp4_add_s3_queries(
      r, url, dir_conf->s3_secret_key, dir_conf->s3_access_key,
      dir_conf->s3_region, dir_conf->s3_security_token, set_url,
      result_text, sizeof result_text);
    if(status != 200)
    {
      ap_log_rerror(APLOG_MARK, APLOG_ERR, status, r,
        "adding S3 signature arguments failed with status %d: %s",
        status, result_text);
      return status;
    }
  }
  else
  {
    // If we need to add headers, check first whether there are S3 auth related
    // query arguments, and bail out in that case. AWS does not accept multiple
    // auth mechanisms simultaneously.
    if(r->args != NULL && (ap_strstr_c(r->args, "AWSAccessKeyId=") != NULL ||
                           ap_strstr_c(r->args, "X-Amz-Signature=") != NULL))
    {
      WARNING_R(r, "Not adding S3 authentication headers, because S3 query "
        "arguments are already set in the request: %s", r->args);
      return DECLINED;
    }

    DEBUG_R(r, "Adding S3 headers, url=%s, args=%s, region=%s",
      url, r->args, dir_conf->s3_region);

    char result_text[256] = { 0 };
    int status = mp4_add_s3_headers(
      r, url, dir_conf->s3_secret_key, dir_conf->s3_access_key,
      dir_conf->s3_region, dir_conf->s3_security_token, set_header_in,
      result_text, sizeof result_text);
    if(status != 200)
    {
      ap_log_rerror(APLOG_MARK, APLOG_ERR, status, r,
        "adding S3 signature headers failed with status %d: %s",
        status, result_text);
      return status;
    }
  }

  TRACE3_R(r, "s3_auth_fixups: leave, DECLINED");
  return DECLINED;
}

static char const* const unified_modules[] =
{
  "mod_smooth_streaming.cpp",
  "mod_unified_remix.cpp",
  "mod_unified_transcode.cpp",
  NULL
};

static void register_hooks(apr_pool_t* pool)
{
  TRACE3_P(pool, "register_hooks: enter");

  // URI-to-filename translation
  ap_hook_translate_name(s3_auth_translate, unified_modules, NULL, APR_HOOK_MIDDLE);

  // at end of request preparation, but before the handler
  ap_hook_fixups(s3_auth_fixups, unified_modules, NULL, APR_HOOK_LAST);

  TRACE3_P(pool, "register_hooks: leave");
}

// End Of File
